--
--      Copyright Cypress Semiconductor Corporation, 1995,1996
--	as an unpublished work.
--

--
-- File contains modules intended for internal use of Warp only
--

library ieee;
use ieee.std_logic_1164.all;

package modpkg_local_ss is

    constant noWarning : boolean := false;
    constant null_vect : std_logic_vector (0 downto 1) := (others => '0');

    function resize (x : std_logic_vector; size : natural) return std_logic_vector;
    function str_to_std_logic_vector (a : string; w : positive) return std_logic_vector;

    function lsb1 (val : string; maxsize : natural) return natural;
    function msb1 (val : string; maxsize : natural) return natural;

    function max (a,b : natural) return natural;

    function to_bv (x: std_logic_vector) return Bit_Vector;
    function from_bv (x: Bit_Vector) return std_logic_vector;

    type bvm		is array (integer range <>,integer range <>) of std_logic;
    type Shift_Type	is (LPM_LOGICAL, LPM_ROTATE, LPM_ARITHMETIC);
    type Repre_Type	is (LPM_SIGNED, LPM_UNSIGNED, LPM_GRAY);
    type Truth_Type	is (LPM_F, LPM_FD, LPM_FR, LPM_FDR);
    type CtDir_Type	is (LPM_NO_DIR, LPM_UP, LPM_DOWN);
    type Arith_Type	is (LPM_NO_TYP, LPM_ADD, LPM_SUB);
    type Fflop_Type	is (LPM_DFF,LPM_TFF);
    type ShDir_Type	is (LPM_LEFT,LPM_RIGHT);
    type goal_type	is (speed,area,combinatorial);
    type port_state_type is (ps_tied_low,ps_tied_high,ps_open,ps_connected);

    signal zero : std_logic := '0' ;
    signal one : std_logic := '1' ;

    component Mmult_bb
	generic(lpm_widthA	: positive;
		lpm_widthB	: positive;
		lpm_widthS	: natural := 0;
		lpm_widthP	: positive;
		lpm_representation : repre_type := lpm_unsigned;
		lpm_hint	: goal_type := speed);
	port(	dataA		: in	bit_vector(lpm_widthA - 1 downto 0);
		dataB		: in	bit_vector(lpm_widthB - 1 downto 0);
		sum		: in	bit_vector(lpm_widthS - 1 downto 0);
		result		: out	bit_vector(lpm_widthP - 1 downto 0));
    end component;

    component Mmultc_bb
	generic(lpm_widthA	: positive;
		lpm_widthB	: positive;
		lpm_widthS	: natural := 0;
		lpm_widthP	: positive;
		lpm_representation : repre_type := lpm_unsigned;
		lpm_hint	: goal_type := speed);
	port(	dataA		: in	bit_vector(lpm_widthA - 1 downto 0);
		dataB		: in	bit_vector(lpm_widthB - 1 downto 0);
		sum		: in	bit_vector(lpm_widthS - 1 downto 0);
		result		: out	bit_vector(lpm_widthP - 1 downto 0));
    end component;

    component Madd_sub_bb
	generic(lpm_width	: positive;
		lpm_representation : repre_type := lpm_unsigned;
		lpm_direction	: arith_type := lpm_no_typ;
		lpm_hint	: goal_type := speed);
	port(	dataA		: in	bit_vector(lpm_width - 1 downto 0);
		dataB		: in	bit_vector(lpm_width - 1 downto 0);
		cin		: in	bit;
		add_sub		: in	bit;
		result		: out	bit_vector(lpm_width - 1 downto 0);
		cout		: out	bit;
		overflow	: out   bit);
    end component;
    component Mcompare_bb
	generic(lpm_width	: positive;
		lpm_representation : repre_type := lpm_unsigned;
		lpm_hint	: goal_type := speed);
	port(	dataA		: in	bit_vector(lpm_width - 1 downto 0);
		dataB		: in	bit_vector(lpm_width - 1 downto 0);
		alb		: out	bit;
		aeb		: out	bit;
		agb		: out	bit;
		ageb		: out   bit;
		aleb		: out   bit;
		aneb		: out   bit);
    end component;
    component cinc
	generic(width	: natural);
	port(	dataA	: in	Bit_Vector;
		sum	: out	Bit_Vector);
    end component;

    component finc
	generic(width	: natural);
	port(	dataA	: in	Bit_Vector;
		cin	: in	Bit;
		sum	: out	Bit_Vector);
    end component;

    component cadd
	generic(width	: natural;
		firstOne: natural;
		lastOne : natural);
	port(	A,K	: in	Bit_Vector;
		sum	: out	Bit_Vector);
    end component;

    component cdec
	generic(width	: natural);
	port(	dataA	: in	Bit_Vector;
		dif	: out	Bit_Vector);
    end component;

    component fdec
	generic(width	: natural);
	port(	dataA	: in	Bit_Vector;
		bin	: in	Bit;
		dif	: out	Bit_Vector);
    end component;

    component csub
	generic(width	: natural;
		firstOne: natural;
		lastOne : natural);
	port(	A,K	: in	Bit_Vector;
		dif	: out	Bit_Vector);
    end component;
end modpkg_local_ss ;

library ieee;
use ieee.std_logic_1164.all;

package body modpkg_local_ss is

    -- Returns first least significant std_logic position less than maxsize
    -- or return maxsize
    function lsb1 (val : string ; maxsize : natural) return natural is
	variable indx : natural := 0 ;
    begin
        for i in val'reverse_range loop
	    IF (indx >= maxsize) OR (val(i) = '1') then
    	        return indx;
    	    end IF;
	    indx := indx + 1 ;
        end loop;
        return maxsize;
    end lsb1;

    -- Returns first most significant std_logic position less than maxsize
    -- or return maxsize
    function msb1 (val : string ; maxsize : natural) return natural is
	variable indx : natural := val'length-1 ;
    begin
        for i in val'range loop
	    IF ((indx < maxsize) AND (val(i) = '1')) then
		return indx;
	    end IF;
	    indx := indx - 1 ;
        end loop;
        return maxsize;
    end msb1;

    function resize (x : std_logic_vector; size : natural) return std_logic_vector is
	variable a : std_logic_vector ((x'length-1) downto 0) := x ;
        variable result : std_logic_vector((size - 1) downto 0) := (others => '0');
    begin
        if (size < 1 ) then return null_vect; end if;
        if a'length < size then
	    result((result'high) downto a'length) := (others => '0');
	    result((a'length - 1) downto 0) := a;
        else
	    result := a((size-1) downto 0);
        end if;
        return result;
    end resize;

    function str_to_std_logic_vector (a : string; w : positive) return std_logic_vector is
	variable result : std_logic_vector ((w-1) downto 0) := (others => '0');
	variable indx : natural ;
    begin
	indx := 0 ;
	for i in a'reverse_range loop
	    if (indx < w) then
		case a(i) is
		when '1' =>
		    result(indx) := '1' ;
		when '0' =>
		    result(indx) := '0' ;
		when others =>
		    Assert (false)
			Report "Invalid char. literal (must be '1' or '0'). Assuming '0'"
			Severity Warning ;
		    result(indx) := '0' ;
		end case ;
	        indx := indx + 1 ;
	    end if ;
	end loop;
	return result;
    end str_to_std_logic_vector;

    function max (a,b : natural) return natural is
    begin
	if (a < b) then return b; else return a; end if;
    end max;

    function to_bv (x: std_logic_vector) return Bit_Vector is
	variable result : Bit_Vector(x'range) ;
    begin
	for i in x'range loop
	    if x(i) = '1' then
		result(i) := '1' ;
	    else
		result(i) := '0' ;
	    end if ;
	end loop ;
	return result ;
    end to_bv ;

    function from_bv (x: Bit_Vector) return std_logic_vector is
	variable result : std_logic_vector(x'range) ;
    begin
	for i in x'range loop
	    if x(i) = '1' then
		result(i) := '1' ;
	    else
		result(i) := '0' ;
	    end if ;
	end loop ;
	return result ;
    end from_bv ;

end modpkg_local_ss ;

--------------------------------------------------------------------------------

--
--	add_2s
--

library ieee;
use ieee.std_logic_1164.all;

use work.modpkg_local_ss.all;

entity warp_add_2s_ss is
    generic (goal : goal_type := speed; signed : Boolean := false);
    port (a,b : in std_logic_vector; s : out std_logic_vector);
end warp_add_2s_ss;

architecture module of warp_add_2s_ss is

    signal dataA,dataB,sum : Bit_Vector ((s'length-1) downto 0);
    signal zero : Bit := '0' ;
    signal one : Bit := '1' ;

begin

    L0:if (a'length < s'length) generate
	dataA <= TO_BV(resize(a,s'length));
    end generate;
    L1:if (a'length = s'length) generate
	dataA <= TO_BV(a);
    end generate;
    L2:if (b'length < s'length) generate
	dataB <= TO_BV(resize(b,s'length));
    end generate;
    L3:if (b'length = s'length) generate
	dataB <= TO_BV(b);
    end generate;

    U0: madd_sub_bb generic map (s'length,lpm_unsigned,lpm_add,goal)
		    port map (dataA,dataB,zero,one,sum,open,open);
    s <= FROM_BV(sum);

end module;

--
--	add_1s1c
--

library ieee;
use ieee.std_logic_1164.all;

use work.modpkg_local_ss.all;

entity warp_add_1s1c_ss is
    generic (c: in string; goal: goal_type := speed; signed: Boolean := false);
    port (a : in std_logic_vector; s : out std_logic_vector);
end warp_add_1s1c_ss;

architecture module of warp_add_1s1c_ss is

    constant lsbOne : natural := lsb1(c,s'length);
    constant msbOne  : natural := msb1(c,s'length);
    signal dataA,dataB,carry,sum : Bit_Vector ((s'length-1) downto 0);
    signal one : Bit := '1' ;

begin

    L0:if (a'length /= s'length) generate
	dataA <= TO_BV(resize(a,s'length));
    end generate;
    L1:if (a'length  = s'length) generate
	dataA <= TO_BV(a);
    end generate;

    dataB <= TO_BV(str_to_std_logic_vector(c,s'length));

    Z0: if (lsbOne >= s'length) generate	-- Const equal to 0 or too big
	s <= FROM_BV(dataA);
    end generate;
    Z1: if (lsbOne <  s'length) generate	-- Const NOT equal to 0
	A0: if (goal /= speed) generate
	    G0: if (lsbOne /= 0) OR (msbOne /= 0) generate
		-- Constant NOT equal to one
	    	G0: for i in dataA'reverse_range generate
		    G1: if (i  = dataA'low) generate
		    	sum(i) <= (dataA(i) XOR dataB(i));
		    	carry(i) <= (dataA(i) AND dataB(i));
		    end generate;
		    G2: if (i /= dataA'low) generate
		        sum(i) <= (dataA(i) XOR dataB(i)) XOR carry(i-1);
		        carry(i) <= (dataA(i) AND dataB(i)) OR
				  (dataA(i) AND carry(i-1)) OR
				  (dataB(i) AND carry(i-1));
		    end generate;
		end generate;
	    end generate;

	    G1: if (lsbOne = 0) AND (msbOne = 0) generate
		-- Const equal to 1
		U0: cinc generic map (s'length) port map (dataA,sum);
	    end generate;
	end generate;

	S0: if (goal = speed) generate
	    G0: if (lsbOne /= 0) OR (msbOne /= 0) generate
		-- Const Not equal to 1
		U0: cadd generic map (s'length, lsbOne, msbOne) 
			port map (dataA,dataB,sum);
	    end generate;

	    G1: if (lsbOne = 0) AND (msbOne = 0) generate
		-- Const equal to 1
		U0: finc generic map (s'length) port map (dataA,one,sum);
	    end generate;
	end generate;
	s <= FROM_BV(sum);
    end generate;
end module;

--
--	sub_2s
--

library ieee;
use ieee.std_logic_1164.all;

use work.modpkg_local_ss.all;

entity warp_sub_2s_ss is
    generic (goal : goal_type := speed; signed : Boolean := false);
    port (a,b : in std_logic_vector; s : out std_logic_vector);
end warp_sub_2s_ss;

architecture module of warp_sub_2s_ss is

    signal dataA,dataB,dif : Bit_Vector ((s'length-1) downto 0);
    signal zero : Bit := '0' ;
    signal one : Bit := '1' ;

begin

    L0:if (a'length < s'length) generate
	dataA <= TO_BV(resize(a,s'length));
    end generate;
    L1:if (a'length = s'length) generate
	dataA <= TO_BV(a);
    end generate;
    L2:if (b'length < s'length) generate
	dataB <= TO_BV(resize(b,s'length));
    end generate;
    L3:if (b'length = s'length) generate
	dataB <= TO_BV(b);
    end generate;

    U0: madd_sub_bb generic map (s'length,lpm_unsigned,lpm_sub,goal)
		    port map (dataA,dataB,one,zero,dif,open,open);
    s <= FROM_BV(dif);

end module;

--
--	sub_1s1c
--

library ieee;
use ieee.std_logic_1164.all;

use work.modpkg_local_ss.all;

entity warp_sub_1s1c_ss is
    generic (c: in string; goal:goal_type := speed; signed : Boolean := false);
    port (a : in std_logic_vector; s : out std_logic_vector);
end warp_sub_1s1c_ss;

architecture module of warp_sub_1s1c_ss is

    constant lsbOne : natural := lsb1(c,s'length);
    constant msbOne  : natural := msb1(c,s'length);
    signal dataA,dataB,borrow,dif : Bit_Vector ((s'length-1) downto 0);
    signal one : Bit := '1' ;

begin

    L0:if (a'length /= s'length) generate
	dataA <= TO_BV(resize(a,s'length));
    end generate;
    L1:if (a'length  = s'length) generate
	dataA <= TO_BV(a);
    end generate;

    dataB <= TO_BV(str_to_std_logic_vector(c,s'length));

    Z0: if (lsbOne >= s'length) generate	-- Const equal to 0 or too big
	s <= FROM_BV(dataA);
    end generate;
    Z1: if (lsbOne <  s'length) generate	-- Const NOT equal to 0
	A0: if (goal /= speed) generate
	    G0: if (lsbOne /= 0) OR (msbOne /= 0) generate
		-- Constant NOT equal to one
	    	G0: for i in dataA'reverse_range generate
		    G1: if (i  = dataA'low) generate
		    	dif(i) <= NOT (NOT dataA(i) XOR dataB(i));
		    	borrow(i) <= (NOT dataA(i) AND dataB(i));
		    end generate;
		    G2: if (i /= dataA'low) generate
		        dif(i) <=NOT(NOT dataA(i) XOR dataB(i)) XOR borrow(i-1);
		        borrow(i) <= (NOT dataA(i) AND dataB(i)) OR
				  (NOT dataA(i) AND borrow(i-1)) OR
				  (dataB(i) AND borrow(i-1));
		    end generate;
		end generate;
	    end generate;

	    G1: if (lsbOne = 0) AND (msbOne = 0) generate
		-- Const equal to 1
		U0: cdec generic map (s'length) port map (dataA,dif);
	    end generate;
	end generate;

	S0: if (goal = speed) generate
	    G0: if (lsbOne /= 0) OR (msbOne /= 0) generate
		-- Const Not equal to 1
		U0: csub generic map (s'length, lsbOne, msbOne) 
			 port map (dataA,dataB,dif);
	    end generate;

	    G1: if (lsbOne = 0) AND (msbOne = 0) generate
		-- Const equal to 1
		U0: fdec generic map (s'length) port map (dataA,one,dif);
	    end generate;
	end generate;
	s <= FROM_BV(dif);
    end generate;

end module;

--
--	sub_1c1s
--

library ieee;
use ieee.std_logic_1164.all;

use work.modpkg_local_ss.all;

entity warp_sub_1c1s_ss is
    generic (c: in string; goal:goal_type := speed; signed : Boolean := false);
    port (b : in std_logic_vector; s : out std_logic_vector);
end warp_sub_1c1s_ss;

architecture module of warp_sub_1c1s_ss is

    signal dataA,dataB,borrow,dif : Bit_Vector ((s'length-1) downto 0);

begin

    L0:if (b'length /= s'length) generate
	dataB <= TO_BV(resize(b,s'length));
    end generate;

    L1:if (b'length  = s'length) generate
	dataB <= TO_BV(b);
    end generate;

    dataA <= TO_BV(str_to_std_logic_vector(c,s'length));

    A0: if (goal /= speed) generate
	G0: for i in dataA'reverse_range generate
	    G1: if (i  = dataA'low) generate
		dif(i) <= NOT(NOT dataA(i) XOR dataB(i));
		borrow(i) <= (NOT dataA(i) AND dataB(i));
	    end generate;
	    G2: if (i /= dataA'low) generate
		dif(i) <= NOT(NOT dataA(i) XOR dataB(i)) XOR borrow(i-1);
		borrow(i) <= (NOT dataA(i) AND dataB(i)) OR
			      (NOT dataA(i) AND borrow(i-1)) OR
			      (dataB(i) AND borrow(i-1));
	    end generate;
	end generate;
	s <= FROM_BV(dif);
    end generate;

    S0: if (goal = speed) generate
	U0: csub generic map (s'length,0,(s'length-1))
		    port map (dataA,dataB,dif);
	s <= FROM_BV(dif);
    end generate;

end module;

--
--	cmp_2s
--

library ieee;
use ieee.std_logic_1164.all;

use work.modpkg_local_ss.all;

entity warp_cmp_2s_ss is
    generic (goal : goal_type := speed; signed : Boolean := false);
    port (a,b : in std_logic_vector; lt,lte,gt,gte,eq,neq: out Boolean);
end warp_cmp_2s_ss;

architecture module of warp_cmp_2s_ss is

    constant size : natural := max(a'length, b'length);
    signal dataA,dataB : Bit_Vector ((size - 1) downto 0);
    signal xlt,xlte,xgt,xgte,xeq,xneq : Bit ;

begin

    L0:if (a'length < b'length) generate
	dataA <= TO_BV(resize(a,b'length));
	dataB <= TO_BV(b);
    end generate;
    L1:if (a'length > b'length) generate
	dataA <= TO_BV(a);
	dataB <= TO_BV(resize(b,a'length));
    end generate;
    L2:if (a'length = b'length) generate
	dataA <= TO_BV(a);
	dataB <= TO_BV(b);
    end generate;

    U0: mcompare_bb generic map (dataA'length,lpm_unsigned,goal)
		    port map (dataA,dataB,xlt,xeq,xgt,xgte,xlte,xneq);
    lt  <= true when  xlt = '1' else false ;
    eq  <= true when  xeq = '1' else false ;
    gt  <= true when  xgt = '1' else false ;
    gte <= true when xgte = '1' else false ;
    lte <= true when xlte = '1' else false ;
    neq <= true when xneq = '1' else false ;

end module;

--
--	cmp_1s1c
--

library ieee;
use ieee.std_logic_1164.all;

use work.modpkg_local_ss.all;

entity warp_cmp_1s1c_ss is
    generic (c : string; goal : goal_type := speed; signed : Boolean := false);
    port (a : in std_logic_vector; lt,lte,gt,gte,eq,neq: out Boolean);
end warp_cmp_1s1c_ss;

architecture module of warp_cmp_1s1c_ss is

    constant size : natural := max(a'length, c'length);
    signal x,b,eqa,lta,gta : std_logic_vector ((size-1) downto 0);

begin

    L0:if (a'length <  size) generate
	x <= resize(a,size);
    end generate;
    L1:if (a'length >= size) generate
	x <= a;
    end generate;
    b <= str_to_std_logic_vector(c,size);

    G0: if ((eq'port_state_att  = ps_connected) OR
	    (neq'port_state_att = ps_connected)) generate
	eqa(size-1) <= NOT (x(size-1) XOR b(size-1));
	G1: for h in (size-2) downto 0 generate
	    eqa(h) <= NOT (x(h) XOR b(h)) AND eqa(h+1);
	end generate;
    end generate;

    G2: if ((lt'port_state_att  = ps_connected) OR
	    (lte'port_state_att = ps_connected) OR
	    (gt'port_state_att  = ps_connected) OR
	    (gte'port_state_att = ps_connected)) generate

	lta(size-1) <= b(size-1) AND NOT (x(size-1) AND b(size-1));
	gta(size-1) <= x(size-1) AND NOT (x(size-1) AND b(size-1));

	G3: for h in (size-2) downto 0 generate
	lta(h) <= ((b(h) AND NOT (x(h) AND b(h))) AND NOT gta(h+1)) OR lta(h+1);
	gta(h) <= ((x(h) AND NOT (x(h) AND b(h))) AND NOT lta(h+1)) OR gta(h+1);
	end generate;
    end generate;

    G4: if (lt'port_state_att  = ps_connected) generate
	lt  <= true when (lta(0) = '1') else false;
    end generate;
    G5: if (lte'port_state_att = ps_connected) generate
	lte <= true when (gta(0) = '0') else false;
    end generate;
    G6: if (gt'port_state_att  = ps_connected) generate
	gt  <= true when (gta(0) = '1') else false;
    end generate;
    G7: if (gte'port_state_att = ps_connected) generate
	gte <= true when (lta(0) = '0') else false;
    end generate;
    G8: if (eq'port_state_att  = ps_connected) generate
	eq  <= true when (eqa(0) = '1') else false;
    end generate;
    G9: if (neq'port_state_att = ps_connected) generate
	neq <= true when (eqa(0) = '0') else false;
    end generate;

end module;

--
--	mul_2s
--

library ieee;
use ieee.std_logic_1164.all;

use work.modpkg_local_ss.all;

entity warp_mul_2s_ss is
    generic (goal : goal_type := speed; signed : Boolean := false);
    port (a,b : in std_logic_vector; p : out std_logic_vector);
end warp_mul_2s_ss;

architecture module of warp_mul_2s_ss is

    constant sum_length : natural :=0;
    signal nosum : Bit_Vector (-1 downto 0);
    signal dataA : Bit_Vector ((a'length-1) downto 0);
    signal dataB : Bit_Vector ((b'length-1) downto 0);
    signal prod  : Bit_Vector ((p'length-1) downto 0);

begin

    dataA <= TO_BV(a);
    dataB <= TO_BV(b);

    U0:mmult_bb generic map(lpm_widthA => a'length,
			    lpm_widthB => b'length,
			    lpm_widthS => sum_length,
			    lpm_widthP => p'length,
			    lpm_representation => lpm_unsigned,
			    lpm_hint => goal)
		 port map (dataA,dataB,nosum,prod);
    p <= FROM_BV(prod((p'length-1) downto 0));

end module;

--
--	mul_1s1c
--

library ieee;
use ieee.std_logic_1164.all;

use work.modpkg_local_ss.all;

entity warp_mul_1s1c_ss is
    generic (c : string; goal : goal_type := speed; signed : Boolean := false);
    port (a : in std_logic_vector; p : out std_logic_vector);
end warp_mul_1s1c_ss;

architecture module of warp_mul_1s1c_ss is

    constant lsbOne : natural := lsb1(c,a'length);
    constant msbOne : natural := msb1(c,a'length);
    signal nosum : Bit_Vector (-1 downto 0);
    signal dataA : Bit_Vector ((a'length-1) downto 0);
    signal dataB : Bit_Vector ((p'length - a'length - 1) downto 0);
    signal prod  : Bit_Vector ((p'length-1) downto 0);
begin

    dataA <= TO_BV(a);

    G0: if (lsbOne < a'length) AND (lsbOne /= msbOne) generate
	-- Constant not equal to zero and not a power of 2
	dataB <= TO_BV(str_to_std_logic_vector(c,dataB'length));
	U0:mmultc_bb generic map(lpm_widthA => dataa'length,
				lpm_widthB => datab'length,
				lpm_widthS => 0,
				lpm_widthP => (dataA'length + dataB'length),
				lpm_representation => lpm_unsigned,
				lpm_hint => goal)
		     port map (dataA,dataB,nosum,prod);
    end generate;

    G1: if (lsbOne < a'length) AND (lsbOne  = msbOne) generate
	-- Constant not equal to zero and is a power of 2
	G2: for i in 0 to (lsbOne-1) generate
	    prod(i) <= '0';
	end generate;
	G3: for i in lsbOne to (lsbOne - 1 + dataA'length) generate
	    prod(i) <= dataA(i - lsbOne);
	end generate;
	G4: for i in (lsbOne + dataA'length) to (p'length - 1) generate
	    prod(i) <= '0';
	end generate;
    end generate;

    G2: if (lsbOne >= a'length) generate	-- Const equal to 0 or too big
	prod <= (others => '0') ;
    end generate ;

    p <= FROM_BV(prod);

end module;
