--
--	Copyright Cypress Semiconductor Corporation, 1994,1995,1996
--	  as an unpublished work.
--
--------------------------------------------------------------------------
--		Components for Generic RTL.				--
--------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

PACKAGE rtlpkg IS

    COMPONENT bufoe		-- Tri-state input/output with feedback or I/O.
	PORT (
	    x	: IN std_logic;		-- Logic input to Buffer.
	    oe	: IN std_logic;		-- Output Enable input.
	    y	: inout std_logic;		-- std_logic output pin.
	    yfb	: OUT std_logic		-- Output feed back as std_logic.
	);
    END COMPONENT;
    attribute atomic_rtl of bufoe : component is bufoe_type;

    COMPONENT dltch		-- Transparent D latch.
	PORT (
	    d	: IN std_logic;		-- D input to latch.
	    e	: IN std_logic;		-- Enabel, tranparent high.
	    q	: OUT std_logic		-- Qutput from latch.
	);
    END COMPONENT;

    COMPONENT dff		-- D type flip flop. (only D and clk input)
	PORT (
	    d	: IN std_logic;		-- D input to register.
	    clk	: IN std_logic;		-- Positive edge Clock.
	    q	: OUT std_logic		-- Output from register.
	);
    END COMPONENT;
    attribute atomic_rtl of dff : component is dff_type;

    COMPONENT xdff		-- X type flip flop. (X1, X2 and clk input)
	PORT (
	    x1	: IN std_logic;		-- First Exclusive OR input.
	    x2	: IN std_logic;		-- Second Exclusive OR input.
	    clk	: IN std_logic;		-- Positive edge Clock.
	    q	: OUT std_logic		-- Qutput from register.
	);
    END COMPONENT;
    attribute atomic_rtl of xdff : component is xdff_type;

    COMPONENT jkff		-- JK type flip flop. ( J, K and clk input)
	PORT (
	    j	: IN std_logic;		-- J input to J-K register.
	    k	: IN std_logic;		-- K input to J-K register.
	    clk	: IN std_logic;		-- Positive edge Clock.
	    q	: OUT std_logic		-- Qutput from register.
	);
    END COMPONENT;

    COMPONENT buf		-- Buffer
	PORT (
	    x	: IN std_logic;		-- Input to Buffer.
	    y	: OUT std_logic		-- Output from Buffer.
	);
    END COMPONENT;
    attribute atomic_rtl of buf : component is buf_type;

    COMPONENT srlch		-- Set, Reset Latch. (S and R input only)
	PORT (
	    s	: IN std_logic;		-- Set input, Async.
	    r	: IN std_logic;		-- Reset input, Async.
	    q	: OUT std_logic		-- Output from Latch.
	);
    END COMPONENT;

    COMPONENT srff		-- Synchronous Set Reset Flip Flop.
	PORT (
	    s	: IN std_logic;		-- Set input, Sync.
	    r	: IN std_logic;		-- Reset inut, Sync.
	    clk	: IN std_logic;		-- Positive edge Clock.
	    q	: OUT std_logic		-- Output from register.
	);
    END COMPONENT;
--    attribute atomic_rtl of srff : component is srff_type;
--	The previous line has been remove for Release 1.1 ONLY.

    COMPONENT dsrff		-- D type flip flop with async. set and reset.
	PORT (
	    d	: IN std_logic;		-- D input to register.
	    s	: IN std_logic;		-- Set input Async.
	    r	: IN std_logic;		-- Reset input Async.
	    clk	: IN std_logic;		-- Positive edge Clock.
	    q	: OUT std_logic		-- Output from register.
	);
    END COMPONENT;
    attribute atomic_rtl of dsrff : component is dsrff_type;

    COMPONENT tsrff		-- T type flip flop with async. set and reset.
	PORT (
	    t	: IN std_logic;		-- T input to register.
	    s	: IN std_logic;		-- Set input Async.
	    r	: IN std_logic;		-- Reset input Async.
	    clk	: IN std_logic;		-- Positive edge Clock.
	    q	: OUT std_logic		-- Output from register.
	);
    END COMPONENT;

    COMPONENT tff		-- T type flip flop. (T and clk input)
	PORT (
	    t	: IN std_logic;		-- Togal input to T type register.
	    clk	: IN std_logic;		-- Positive edge Clock.
	    q	: OUT std_logic		-- Output from register.
	);
    END COMPONENT;

    COMPONENT xbuf		-- Xor primitive gate
	PORT (
	    x1	: IN std_logic;		-- Xor input.
	    x2	: IN std_logic;		-- Xor input.
	    q	: OUT std_logic		-- Xor output.
	);
    END COMPONENT;
    attribute atomic_rtl of xbuf : component is xbuf_type;

    COMPONENT triout		-- Tri-state output with no feedback or I/O.
	PORT (
	    x	: IN std_logic;		-- Logic input to Buffer.
	    oe	: IN std_logic;		-- Output Enable input.
	    y	: OUT std_logic		-- std_logic output pin.
	);
    END COMPONENT;
    attribute atomic_rtl of triout : component is triout_type;

-- C331 specials

    component ir_C331
	port(
	    d	: in std_logic;		-- D input to input register.
	    ar_i	: in std_logic;	-- async reset, input reg.
	    as_i	: in std_logic;	-- async set, input reg.
	    clk_i	: in std_logic;	-- Positive edge, input reg, clock.
	    q_i	: out std_logic		-- Output from input register.
	);
    end component;
    attribute atomic_rtl of ir_C331 : component is ir_C331_type;

END rtlpkg;

--------------------------------------------------------------------------
--		Enties for Generic RTL.					--
--------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

USE WORK.rtlpkg.ALL;
ENTITY dltch IS
    PORT (
	d	: IN std_logic;		-- D input to latch.
	e	: IN std_logic;		-- Enabel, tranparent high.
	q	: OUT std_logic		-- Qutput from latch.
    );
END dltch;

library ieee;
use ieee.std_logic_1164.all;

USE WORK.rtlpkg.ALL;
ENTITY jkff IS
    PORT (
	j	: IN std_logic;		-- J input to J-K register.
	k	: IN std_logic;		-- K input to J-K register.
	clk	: IN std_logic;		-- Positive edge Clock.
	q	: OUT std_logic		-- Qutput from register.
    );
END jkff;

library ieee;
use ieee.std_logic_1164.all;

USE WORK.rtlpkg.ALL;
ENTITY srlch IS
    PORT (
	s	: IN std_logic;		-- Set input, Async.
	r	: IN std_logic;		-- Reset input, Async.
	q	: OUT std_logic		-- Output from Latch.
    );
END srlch;

library ieee;
use ieee.std_logic_1164.all;

USE WORK.rtlpkg.ALL;
ENTITY tsrff IS	
    PORT (t	: IN std_logic;
	s	: IN std_logic;
	r	: IN std_logic;
	clk	: IN std_logic;
	q	: OUT std_logic);
END tsrff;

library ieee;
use ieee.std_logic_1164.all;

USE WORK.rtlpkg.ALL;
ENTITY tff IS
    PORT (
	t	: IN std_logic;		-- Togal input to T type register.
	clk	: IN std_logic;		-- Positive edge Clock.
	q	: OUT std_logic		-- Output from register.
    );
END tff;

--------------------------------------------------------------------------
--		Architectures for Generic RTL				--
--------------------------------------------------------------------------

--
-- FOR Release 1.1 ONLY.

library ieee;
use ieee.std_logic_1164.all;

USE WORK.rtlpkg.ALL;
ENTITY srff IS
    PORT (
	s	: IN std_logic;		-- First Exclusive OR input.
	r	: IN std_logic;		-- Second Exclusive OR input.
	clk	: IN std_logic;		-- Positive edge Clock.
	q	: OUT std_logic		-- Qutput from register.
    );
END srff;

ARCHITECTURE archsrff OF srff IS
    SIGNAL dx,qx : std_logic;
BEGIN
    dx	<= (s AND NOT r) OR (NOT r AND qx);
    u1 :dff port map(dx,clk,qx);
    q <= qx;
END archsrff;
-- FOR Release 1.1 ONLY.
--

ARCHITECTURE archdltch OF dltch IS
    SIGNAL qx : std_logic;
BEGIN
    qx	<= (d and e) or (d and qx) or (not(e) and qx);
    q	<= qx;
END archdltch;

ARCHITECTURE archjkff OF jkff IS
    SIGNAL qx, dx : std_logic;
BEGIN
    dx	<= (j and not(qx)) or (not(k) and qx);
    q	<= qx;
    d1:dff port map(dx,clk,qx);
END archjkff;
	
ARCHITECTURE archsrl OF srlch IS
    SIGNAL qx : std_logic;
BEGIN
    qx	<= (s and not(r)) or (qx and not(r));
    q	<= qx;
END archsrl;

ARCHITECTURE archTSRFF of tsrff IS
   signal dtmp,qtmp:std_logic;
BEGIN
	dtmp <= qtmp xor t;
	x:dsrff port map(dtmp,s,r,clk,qtmp);
	q <= qtmp;
END archTSRFF;

ARCHITECTURE archtff OF tff IS
    SIGNAL qx, dx : std_logic;
BEGIN
    dx	<= t xor qx;
    q	<= qx;
    d:dff port map(dx,clk,qx);
END archtff;

library ieee;
use ieee.std_logic_1164.all;

use work.rtlpkg.all;
ENTITY xor2 IS
    port(
	a,b     : in std_logic;
	q       : out std_logic);
END xor2;
			 
ARCHITECTURE archXOR2 OF xor2 IS
begin
    u1 : xbuf port map (a, b, q);
END archXOR2;

--      Here is the device entity.

library ieee;
use ieee.std_logic_1164.all;

ENTITY C331 IS
PORT(
	pin1  :IN	std_logic;
	pin2  :IN	std_logic;
	pin3  :IN	std_logic;
	pin4  :in	std_logic;
	pin5  :in	std_logic;
	pin6  :in	std_logic;
	pin7  :in	std_logic;
	pin9  :in	std_logic;
	pin10 :in	std_logic;
	pin11 :IN	std_logic;
	pin13 :IN	std_logic;
	pin12 :IN	std_logic;
	pin14 :in	std_logic;
	pin15 :INOUT	std_logic;
	pin16 :INOUT	std_logic;
	pin17 :INOUT	std_logic;
	pin18 :INOUT	std_logic;
	pin19 :INOUT	std_logic;
	pin20 :INOUT	std_logic;
	pin23 :INOUT	std_logic;
	pin24 :INOUT	std_logic;
	pin25 :INOUT	std_logic;
	pin26 :INOUT	std_logic;
	pin27 :INOUT	std_logic;
	pin28 :INOUT	std_logic);
END C331;
